/* lav.c -- read an alignment file in lav or concise format, and
 *          perform a variety of queries on it
 *
 *  Entry points:
 *
 *	int get_align (char *filename, Segment **seg);
 *		Reads alignment information from an lav or concise file
 *		into a new array of Segment structs, and returns the size
 *		of the allocated array for passing to free_align() later
 *		(this size might be slightly larger than the actual amount
 *		of data read).  For an lav file, a Segment is created for
 *		each contig block, each local alignment, and each gap-free
 *		segment, which are interleaved in the order they appear in
 *		the file (i.e., the information in a contig entry applies
 *		to all of the other entries that follow it up to the next
 *		contig entry, and similarly for local alignment entries).
 *		The three types of entries use different subsets of the
 *		available fields; see the struct declaration comments in
 *		lav.h for details.  Note that the endpoint values for local
 *		alignments and gap-free segments are adjusted to refer to
 *		the user's original full sequence files, rather than the
 *		raw (possibly offset or reversed) coordinates produced by
 *		blastz.  For a concise file the array is similar, except
 *		that the data fields for local alignment entries are empty,
 *		and endpoint values for contig entries (which control the
 *		raw -> user coordinate adjustment for lav files) are not
 *		available.  It is assumed that endpoint values in the
 *		concise file have already been adjusted to user coordinates.
 *
 *	void free_align (Segment *seg, int size);
 *		Frees the array created by get_align().  Be sure to give
 *		the full allocated size returned by get_align(), not the
 *		number of entries actually filled.
 *
 *	int find_seg (Segment *seg, int pos, bool inseq2, char *contig,
 *	              int from_index, bool *gap);
 *		Finds the first gap-free segment containing the given
 *		position.  Positions falling between gap-free segments
 *		within a local alignment are rounded to the lower-indexed
 *		segment; *gap is set to indicate if this occurred.  Specify
 *		a contig header (or prefix) in the second sequence to
 *		restrict the segments used, or use NULL to mean "any contig".
 *		Searching begins at from_index (normally 0) and goes up
 *		from there; this enables repeated calls to find additional
 *		matches.  However, it is the RESPONSIBILITY OF THE CALLER
 *		to ensure that from_index is a legitimate index for the seg
 *		array, i.e., not out-of-bounds.  If no matching segment is
 *		found, the return value is -1.
 *
 *	int convert_pos (Segment *seg, int i, int pos, bool inverse, bool gap);
 *		Converts a position from top sequence to bottom, or the
 *		inverse, using the segment at the specified index.  It is
 *		the RESPONSIBILITY OF THE CALLER to ensure that i is not
 *		out-of-bounds.  Setting gap to true indicates that the
 *		match fell between this segment and the next, and should
 *		be rounded to the end of this segment.  Typically i and
 *		gap are the results from a prior call to find_seg().  If
 *		i < 0 (no segment found) then the return value is 0.
 *
 *	bool has_multi_contig (Segment **seg);
 *		Checks whether the second sequence contained multiple
 *		contigs (reverse complement doesn't count).
 *
 *	char *get_hdr2 (Segment *seg, int i);
 *		Returns a pointer to the second-sequence header stored in
 *		the contig entry associated with the entry at index i.
 *
 *	bool same_local (Segment *seg, int i1, int i2);
 *		Determines whether two segment entries are in the same
 *		local alignment.
 *
 * $Revision: 1.21 $$Date: 2001/10/24 22:38:25 $
 */

#include "lav.h"    /* typedef for Segment */
#include "util.h"

#define LAV    "#:lav"
#define H2PRE  "Aligned to \""
#define H2SUF  "\":\n"
#define REV1   "-\""
#define REV2   " (reverse complement)"
#define LOCAL  ":"
#define ARROW  "<-->"

enum { BUFSIZE = 1000 };

static int read_lav (FILE *fp, Segment **seg);
static int read_concise (FILE *fp, Segment **seg);
static void get_seqends (char *buf, int *start, int *end);
static char *get_header_lav (char *buf, bool reverse);
static void get_header_concise (char *buf, char **hdr, bool *reverse);
static void adjust_coords (Segment *seg);
static void spread_reverse (Segment *seg);

int get_align (char *filename, Segment **seg) {
	FILE *fp;
	char *stat;                        /* read status */
	char buf [BUFSIZE];
	char *p;
	int size = 0;

	fp = ckopen (filename, "r");
	while ((stat = fgets (buf, BUFSIZE, fp)) && !*skip_ws (buf))
		;
	if (!stat)
		fatalf ("\"%s\" is empty.", filename);

	p = skip_ws (buf);
	if (starts (p, LAV))
		size = read_lav (fp, seg);
	else if (starts (p, H2PRE))
		size = read_concise (fp, seg);
	else
		fatalf ("\"%s\" is not an alignment file.", filename);

	fclose (fp);
	return size;
}

/* read an alignment file in lav format */
static int read_lav (FILE *fp, Segment **seg) {
	char buf [BUFSIZE];
	char *p;
	int nseg = 0;
	int i;

	/* allocate Segment array */
	while (fgets (buf, BUFSIZE, fp)) {
		p = skip_ws (buf);
		if (starts (p, "s {") || starts (p, "a {") || starts (p, "l "))
			nseg++;    /* this estimate may be slightly high */
	}
	nseg++;    /* end marker */
	*seg = (Segment *) ckallocz (nseg * sizeof (Segment));
	rewind (fp);

	/* read alignment file into array */
	i = -1;
	while (fgets (buf, BUFSIZE, fp)) {
		if (starts (skip_ws (buf), "d {")) {
			while (fgets (buf, BUFSIZE, fp) && !starts (skip_ws (buf), "}"))
				;    /* skip lines that might start with s, b, e, or l */
		}
		else if (starts (skip_ws (buf), "s {")) {
			i++;
			(*seg)[i].type = 's';
			if (!fgets (buf, BUFSIZE, fp))
				fatalf ("Unexpected end of lav file.");
			get_seqends (buf, &((*seg)[i].x0), &((*seg)[i].x1));
			if (!fgets (buf, BUFSIZE, fp))
				fatalf ("Unexpected end of lav file.");
			get_seqends (buf, &((*seg)[i].y0), &((*seg)[i].y1));
			(*seg)[i].reverse = (strstr (buf, REV1) != NULL);
			(*seg)[i].hdr2 = NULL;   /* in case there's no "h {" later */
		}
		else if (starts (skip_ws (buf), "h {")) {
			if (i < 0 || (*seg)[i].type != 's')
				fatalf ("Unexpected line in lav file:\n%s", buf);
			if (!fgets (buf, BUFSIZE, fp) || !fgets (buf, BUFSIZE, fp))
				fatalf ("Unexpected end of lav file.");
			(*seg)[i].hdr2 = get_header_lav (buf, (*seg)[i].reverse);
		}
		else if (starts (skip_ws (buf), "a {")) {
			i++;
			(*seg)[i].type = 'x';    /* partial */
		}
		else if (starts (skip_ws (buf), "s")) {
			if (i < 0 || (*seg)[i].type != 'x')
				fatalf ("Unexpected line in lav file:\n%s", buf);
			if (sscanf (buf, " s %d",
			            &((*seg)[i].score)) != 1)
				fatalf ("Bad line in lav file:\n%s", buf);
		}
		else if (starts (skip_ws (buf), "b")) {
			if (i < 0 || (*seg)[i].type != 'x')
				fatalf ("Unexpected line in lav file:\n%s", buf);
			if (sscanf (buf, " b %d %d",
			            &((*seg)[i].x0), &((*seg)[i].y0)) != 2)
				fatalf ("Bad line in lav file:\n%s", buf);
		}
		else if (starts (skip_ws (buf), "e")) {
			if (i < 0 || (*seg)[i].type != 'x')
				fatalf ("Unexpected line in lav file:\n%s", buf);
			if (sscanf (buf, " e %d %d",
			            &((*seg)[i].x1), &((*seg)[i].y1)) != 2)
				fatalf ("Bad line in lav file:\n%s", buf);
			(*seg)[i].type = 'a';
		}
		else if (starts (skip_ws (buf), "l")) {
			i++;
			(*seg)[i].type = 'l';
			if (sscanf (buf, " l %d %d %d %d %d",
			            &((*seg)[i].x0), &((*seg)[i].y0),
			            &((*seg)[i].x1), &((*seg)[i].y1),
			            &((*seg)[i].score)) != 5)
				fatalf ("Bad line in lav file:\n%s", buf);
		}
	}
	(*seg)[++i].type = 'z';
	adjust_coords (*seg);
	spread_reverse (*seg);
	return nseg;
}

/* read an alignment file in concise format */
static int read_concise (FILE *fp, Segment **seg) {
	char buf [BUFSIZE];
	char *p;
	int nseg = 0;
	int i;

	/* allocate Segment array */
	rewind (fp);
	while (fgets (buf, BUFSIZE, fp)) {
		p = skip_ws (buf);
		if (starts (p, H2PRE) || starts (p, LOCAL) || strstr (buf, ARROW))
			nseg++;
	}
	nseg++;    /* end marker */
	*seg = (Segment *) ckallocz (nseg * sizeof (Segment));
	rewind (fp);

	/* read alignment file into array */
	i = -1;
	while (fgets (buf, BUFSIZE, fp)) {
		p = skip_ws (buf);
		if (starts (p, H2PRE)) {
			i++;
			(*seg)[i].type = 's';
			(*seg)[i].x0 = (*seg)[i].x1 = (*seg)[i].y0 = (*seg)[i].y1 = 0;
			get_header_concise (buf, &((*seg)[i].hdr2), &((*seg)[i].reverse));
		}
		else if (starts (p, LOCAL)) {
			i++;
			(*seg)[i].type = 'a';
			(*seg)[i].x0 = (*seg)[i].x1 = (*seg)[i].y0 = (*seg)[i].y1 = 0;
			(*seg)[i].score = 0;
		}
		else if (strstr (buf, ARROW)) {
			i++;
			(*seg)[i].type = 'l';
			if (sscanf (buf, "%d-%d <--> %d-%d %d%%",
			            &((*seg)[i].x0), &((*seg)[i].x1),
			            &((*seg)[i].y0), &((*seg)[i].y1),
			            &((*seg)[i].score)) != 5)
				fatalf ("Bad line in concise file:\n%s", buf);
		}
	}
	(*seg)[++i].type = 'z';
	spread_reverse (*seg);
	return nseg;
}

/* extract sequence endpoints from line in "s {" stanza */
static void get_seqends (char *buf, int *start, int *end) {
	if (sscanf (buf, " \"%*[^\"]\" %d %d", start, end) != 2)
		fatalf ("Bad sequence spec in lav file:\n%s", buf);
}

/* extract contig header from line in "h {" stanza; check reverse status */
static char *get_header_lav (char *buf, bool reverse) {
	char buf2 [BUFSIZE];
	char *p;

	if (sscanf (buf, " \"%[^\"]\"", buf2) != 1)
		fatalf ("Bad contig header in lav file:\n%s", buf);
	p = strstr (buf2, REV2);
	if ((!p && reverse) || (p && !reverse))
		fatalf ("Reverse complement indicators do not agree:\n%s", buf);
	if (p)
		*p = '\0';
	return copy_string (buf2);
}

/* extract contig header & reverse status from section line in concise file */
static void get_header_concise (char *buf, char **hdr, bool *reverse) {
	char *p;

	p = strstr (buf, REV2);
	*reverse = (p != NULL);
	p = (p ? p-1 : strstr (buf, H2SUF));
	if (!p || *p != '\"')    /* closing quote inserted by concise-align */
		fatalf ("Bad header format in concise file:\n%s", buf);
	*p = '\0';
	*hdr = copy_string (skip_ws (buf) + strlen (H2PRE));
}

/* convert endpoints from raw blastz coords to user's original sequence */
static void adjust_coords (Segment *seg) {
	int sx0 = 0, sx1 = 0, sy0 = 0, sy1 = 0;    /* seq endpoints */
	bool reverse = 0;
	int i;

	for (i = 0; seg[i].type != 'z'; i++) {
		if (seg[i].type == 's') {
			sx0 = seg[i].x0;
			sx1 = seg[i].x1;
			sy0 = seg[i].y0;
			sy1 = seg[i].y1;
			reverse = seg[i].reverse;
		}
		else {
			seg[i].x0 = seg[i].x0 + sx0 - 1;
			seg[i].x1 = seg[i].x1 + sx0 - 1;
			if (reverse) {
				seg[i].y0 = sy1 + 1 - seg[i].y0;
				seg[i].y1 = sy1 + 1 - seg[i].y1;
			}
			else {
				seg[i].y0 = seg[i].y0 + sy0 - 1;
				seg[i].y1 = seg[i].y1 + sy0 - 1;
			}
		}
	}
}

/* copy reverse values from 's' entries to 'a' and 'l', for convenience */
static void spread_reverse (Segment *seg) {
	bool reverse = 0;
	int i;

	for (i = 0; seg[i].type != 'z'; i++) {
		if (seg[i].type == 's')
			reverse = seg[i].reverse;
		else
			seg[i].reverse = reverse;
	}
}

void free_align (Segment *seg, int size) {
	int i;
	char *p;

	for (i = 0; i < size; i++) {
		if (seg[i].type == 's' && (p = seg[i].hdr2))
			free (p);
	}
	free (seg);
}

int find_seg (Segment *seg, int pos, bool inseq2, char *contig,
              int from_index, bool *gap) {
	int i;
	char type;
	char *hdr2;
	int start, end, last = 0, key;

	/* caller is responsible for ensuring that from_index is valid */
	hdr2 = get_hdr2 (seg, from_index);

	for (i = from_index; (type = seg[i].type) != 'z'; i++) {
		switch (type) {
		case 's':
			hdr2 = seg[i].hdr2;
			break;
		case 'a':
			last = 0;
			break;
		case 'l':
			if (contig && !starts (hdr2, contig))
				continue;
			start = (inseq2 ? seg[i].y0 : seg[i].x0);
			end   = (inseq2 ? seg[i].y1 : seg[i].x1);
			key = pos;
			if (inseq2 && seg[i].reverse) {
				/* fix comparisons */
				start = -start; end = -end; key = -key;
			}
			if (last && last < key && key < start) {
				*gap = 1;
				return i-1;
			}
			else if (start <= key && key <= end) {
				*gap = 0;
				return i;
			}
			else
				last = end;    /* no need to reconstruct this
				                  for from_index > 0 */
			break;
		default:
			fatalf ("Internal error: invalid segment type '%c'.", type);
		}
	}
	return -1;
}

int convert_pos (Segment *seg, int i, int pos, bool inverse, bool gap) {
	if (i < 0)
		return 0;
	else if (gap)
		return (inverse ? seg[i].x1 : seg[i].y1);
	else if (inverse)
		return seg[i].x0 + (seg[i].reverse ? -1 : +1) * (pos - seg[i].y0);
	else
		return seg[i].y0 + (seg[i].reverse ? -1 : +1) * (pos - seg[i].x0);
}

bool seg_overlaps (Segment *seg, int i, int start, int end, bool inseq2) {
	int a0, a1;

	a0 = (inseq2 ? seg[i].y0 : seg[i].x0);
	a1 = (inseq2 ? seg[i].y1 : seg[i].x1);
	return overlaps (a0, a1, start, end);
}

bool has_multi_contig (Segment *seg) {
	char *hdr2 = NULL;
	int i;

	for (i = 0; seg[i].type != 'z'; i++) {
		if (seg[i].type == 's') {
			if (!hdr2)
				hdr2 = seg[i].hdr2;
			else if (!strsame (seg[i].hdr2, hdr2))
				return 1;
		}
	}
	return 0;
}

char *get_hdr2 (Segment *seg, int i) {
	for ( ; i >= 0 && seg[i].type != 's'; i--)
		;
	return (i >= 0 ? seg[i].hdr2 : NULL);
}

bool same_local (Segment *seg, int i1, int i2) {
	int j;

	if (i1 > i2) {
		swap (&i1, &i2);
	}
	for (j = i1; j <= i2; j++) {
		if (seg[j].type != 'l')
			return 0;
	}
	return 1;
}

